def my_awesome_func(num):
    """
    A few things to notice:
    1. Any imports must be done inside the function.
    2. Return either native types or numpy scalars. Do not return non-scalar numpy values.
    """
    import numpy

    return numpy.multiply(num, 2.0)


def get_simulation_name(context):
    return context["simulation"]


def get_sid(context):
    return context["sid"]



from .reader_mcap import BagReaderMcap
from .reader_sqlite import BagReaderSQL

class BagReader():
    def read_messages(self, input_bag, simulation_run_id):
        '''        
        return: 
            StringIO
        '''
        postfix = input_bag.split('.')[-1]
        
        br = None
        if (postfix == 'mcap'):
            br = BagReaderMcap()            
        elif( postfix == 'db3'):
            br = BagReaderSQL()            
        else:
            raise NotImplementedError

        return br.read_messages(input_bag, simulation_run_id)
        


import argparse
import importlib_resources
from bin.cli_impl import *
from rich_argparse import RichHelpFormatter
from rich import print, inspect, print_json
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich_argparse import RichHelpFormatter
from rich.traceback import install
from bin import __version__ as citros_version

install()


# citros simulation run
def parser_simulation_run(subparser, epilog=None):
    description_path = "simulation/run.md"
    help = "citros data create section"
    parser = subparser.add_parser(
        "run",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=simulation_run)

    return parser


# citros simulation list
def parser_simulation_list(subparser, epilog=None):
    description_path = "simulation/list.md"
    help = "citros data status section"
    parser = subparser.add_parser(
        "list",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=simulation_list)
    return parser


# citros simulation
def parser_simulation(subparser, epilog=None):
    description_path = "simulation.md"
    help = "simulation section"
    parser = subparser.add_parser(
        "simulation",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
        aliases=["sim"],
    )
    parser.add_argument("-n", "--name", default=None, help="name of simulation")
    parser.add_argument("-m", "--match", default=None, help="match simulation pattern")

    subparser = parser.add_subparsers(dest="type")
    # simulation run/list
    parser_simulation_run(subparser, epilog=epilog)
    parser_simulation_list(subparser, epilog=epilog)

    return parser



import argparse
import importlib_resources
from bin.cli_impl import *
from rich_argparse import RichHelpFormatter
from rich import print, inspect, print_json
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich_argparse import RichHelpFormatter
from rich.traceback import install
from bin import __version__ as citros_version

install()


def parser_init(main_sub, epilog=None):
    parser = main_sub.add_parser(
        "init",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath("init.md"), "r"
                ).read()
            ),
            subtitle=f"[{citros_version}]",
        ),
        epilog=epilog,
        help="init section",
        formatter_class=RichHelpFormatter,
    )
    parser.add_argument(
        "-dir", "--dir", default=".", help="The working dir of the project"
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=init)



import os
import glob
import json
import logging
from pathlib import Path
from datetime import datetime
from rich import print, inspect, print_json
from rich.logging import RichHandler

from citros.utils import get_user_git_info

from .simulation import Simulation
from .logger import get_logger, shutdown_log

from .batch_uploader import BatchUploader


class NoBatchFoundException(Exception):
    def __init__(self, message="No batch found."):
        super().__init__(message)


# * create new batch for running simulations
# to be able to run simulation Batch needs to have a simulation
# Batch(simulation : Simulation)


# * create new batch for reading data from previous runs
# To be able to interact with recorded batches
# Batch(path: Path, index: int) # path data/simulation/batch,
# index = -1 will get the latest batch run from this dir
# index = n will get the n's batch run
class Batch(BatchUploader):
    def __exit__(self):
        self.log.debug(
            f"{self.__class__.__name__}.__exit__()",
        )
        shutdown_log()

    def __init__(
        self,
        root,  # the base recordings dir
        simulation,  # if type(simulation) str then it is the name of the simulation if type(simulation) Simulation then it is the simulation object
        name: str = "citros",
        message: str = "CITROS is AWESOME!!!",
        index: int = -1,  # default to take the last version of a runs
        version=None,
        log=None,
        debug=False,
        verbose=False,
    ):
        self.debug = debug
        self.verbose = verbose

        if root is None:
            raise Exception("Error: root dir is None, batch needs is to operate")
        if simulation is None:
            raise Exception("Error: simulation is None, batch needs is to operate")
        if type(simulation) is not str and not Simulation:
            raise Exception("Error: simulation is not a string or Simulation object")

        self.root = root
        self.simulation = simulation
        self.name = name
        self.message = message
        self.version = version
        self.index = index

        self.simulation_name = (
            simulation if type(simulation) is str else simulation.name
        )

        self.batch_dir = Path(root) / self.simulation_name / name

        # get version
        if not version:  # no version specified
            if type(simulation) is Simulation:  # create new batch
                self.version = datetime.today().strftime("%Y%m%d%H%M%S")
            else:
                versions = sorted(glob.glob(f"{str(self.batch_dir)}/*/"))
                # get version from index
                version_path = versions[self.index]
                if version_path.endswith("/"):
                    version_path = version_path[:-1]

                self.version = version_path.split("/")[-1]

        # print(f"{self.simulation_name} / {name} / {self.version}")
        self.batch_dir = Path(root) / self.simulation_name / name / self.version

        self._init_log(log)

        self.log.debug(f"{self.__class__.__name__}.init()")
        self.log.debug(f"self.batch_dir:{str(self.batch_dir)}")

        self.data = {}

        # when simulation is a string then we are creating loading a batch from a path
        if type(simulation) is str:
            self._load()
        # when simulation is a Simulation then we are creating new batch and starting a simulations
        else:
            self._new()

        self._validate()

    def __str__(self):
        # print_json(data=self.data)
        return json.dumps(self.data, indent=4)

    def __getitem__(self, key):
        """get element from object

        Args:
            key (str): the element key

        Returns:
            str: the element value
        """
        return self.data[key]

    def get(self, key, default=None):
        """get element from object

        Args:
            key (str): the element key

        Returns:
            str: the element value
        """
        return self.data.get(key, default)

    def __setitem__(self, key, newvalue):
        self.data[key] = newvalue
        self._save()

    ###################
    ##### private #####
    ###################
    def _init_log(self, log=None):
        self.log = log
        if self.log is None:
            if type(self.simulation) is Simulation:  # creating new
                self.batch_dir.mkdir(parents=True, exist_ok=True)
                log_dir = self.batch_dir
            else:
                Path.home().joinpath(".citros/logs").mkdir(parents=True, exist_ok=True)
                log_dir = Path.home().joinpath(".citros/logs")

            self.log = get_logger(
                __name__,
                log_level=os.environ.get("LOGLEVEL", "DEBUG" if self.debug else "INFO"),
                log_file=str(log_dir / "citros.log"),
                verbose=self.verbose,
            )

    # verify that the batch folder is ok:
    # - all json is correct.
    # - all files is intact.
    # - if files is signed check all signings (sha)
    def _validate(self):
        # TODO[critical]: add validations
        return True, None

    def _new(self):
        self.log.debug(
            f"{self.__class__.__name__}._new()",
        )
        self.batch_dir.mkdir(parents=True, exist_ok=True)

        commit, branch = get_user_git_info(self.simulation.root)

        # inspect(self.simulation)
        self.data = {
            "simulation": self.simulation.name,
            "name": self.name,
            "message": self.message,
            "gpu": self.simulation["GPU"],
            "cpu": self.simulation["CPU"],
            "memory": self.simulation["MEM"],
            "timeout": self.simulation["timeout"],
            "commit": commit,
            "branch": branch,
            "storage_type": self.simulation["storage_type"],  # SQLITE, MCAP
            # will be filled at runtime
            "completions": "",
            "status": "",
            "metadata": "",
            "created_at": datetime.today().strftime("%Y-%m-%d %H:%M:%S"),
            "updated_at": datetime.today().strftime("%Y-%m-%d %H:%M:%S"),
        }

        self._save()

    def _save(self):
        self.log.debug(
            f"{self.__class__.__name__}._save()",
        )
        self.data["updated_at"]: datetime.today().strftime("%Y-%m-%d %H:%M:%S")

        with open(self.path(), "w") as file:
            json.dump(self.data, file, indent=4, sort_keys=True)

    def _load(self):
        self.log.debug(
            f"{self.__class__.__name__}._load()",
        )
        if not self.batch_dir.exists():
            raise NoBatchFoundException(f'No batch fount at "{self.batch_dir}"')

        batch_info = self.path()

        self.log.debug(f"loading version: {batch_info}")

        try:
            with open(Path(batch_info), "r") as file:
                batch_run = json.load(file)

                self.data.update(batch_run)
        except FileNotFoundError as e:
            self.log.error(f"no file for {batch_info}")
        except Exception as e:
            self.log.exception(e, self.data)

    ###################
    ##### public ######
    ###################
    def path(self):
        """return the full path to the current main file.

        default to ".citros/project.json"

        Returns:
            str: the full path to the current main file.
        """
        # versions = sorted(glob.glob(f"{str(self.batch_dir)}/*"))
        # batch_version = versions[self.index]

        return self.batch_dir / "info.json"

    def simulation_run(
        self,
        sid: int,
        ros_domain_id: int = None,
        trace_context: str = None,
    ):
        self.log.debug(f"{self.__class__.__name__}.simulation_run()")
        self["status"] = "RUNNING"
        sim_dir = self.batch_dir / str(sid)
        # create log that will write to the simulation dir.
        ret = self.simulation.run(
            sim_dir, sid=sid, trace_context=trace_context, ros_domain_id=ros_domain_id
        )

        self.log.debug(f"{self.__class__.__name__}.run(): ret {ret}")
        return ret

    def run(
        self,
        completions: int = 1,
        sid: int = -1,  # the run id to complete. -1 -> run all
        ros_domain_id: int = None,
        trace_context: str = None,
    ):
        self.log.debug(f"{self.__class__.__name__}.run()")
        self["completions"] = completions

        if sid != -1:
            self.simulation_run(sid, ros_domain_id, trace_context)
            return
        for i in range(int(completions)):
            self.log.debug(f"run {i}")
            self.simulation_run(i, ros_domain_id, trace_context)



import numpy as np
import matplotlib.pyplot as plt

from citros import CitrosDB, Validation
from decouple import config
'''
tests for error_analysis module

just type:

pytest

in terminal in this folder
'''
table_name = 'test_table'

if config('TEST_ENV', None) == 'github':
    citros = CitrosDB(simulation = 'public',
                      batch = table_name,
                      host=config('POSTGRES_HOST'),
                      user=config('POSTGRES_USER'),
                      password=config('POSTGRES_PASSWORD'),
                      database=config('POSTGRES_DB'),
                      port = config('POSTGRES_PORT'))
else:
    citros = CitrosDB(simulation = 'public',
                      batch = table_name)

def test_std_bound_test():
    df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.x', 'data.x.x_1','data.x.x_2','data.x.x_3','data.time'])

    V = Validation(df, data_label = ['data.x.x_1'], param_label = 'data.time', 
                      method = 'bin', num = 50, units = 'm')
    V3 = Validation(df, data_label = ['data.x.x_1', 'data.x.x_2','data.x.x_3'], param_label = 'data.time', 
                        method = 'scale', num = 50, units = 'm')
    log, table, fig = V3.std_bound_test(limits = [0.25, 0.28, [-20, 10]], n_std = 3)
    
    assert (log['test_param']['limits'][0] - 0.25) < 0.001, "std_bound_test: ['test_param']['limits'][0] is wrong"
    assert (log['test_param']['limits'][2][1] - 10) < 0.001, "std_bound_test: ['test_param']['limits'][2][1] is wrong"
    assert log['data.x.x_1']['passed'] == True, "std_bound_test: ['data.x.x_1']['passed'] is wrong"
    assert log['data.x.x_3']['passed'] == False, "std_bound_test: ['data.x.x_3']['passed'] is wrong"
    assert abs(log['data.x.x_1']['pass_rate'] - 1.0) < 0.001, "std_bound_test: ['data.x.x_1']['pass_rate'] is wrong"
    assert abs(log['data.x.x_3']['pass_rate']) < 0.001, "std_bound_test: ['data.x.x_3']['pass_rate'] is wrong"

    assert table['data.x.x_3'].any() == False, "std_bound_test: table['data.x.x_3'] is wrong"
    assert table['data.x.x_1'].all() == True, "std_bound_test: table['data.x.x_1'] is wrong"
    assert len(fig.axes) == 3, "std_bound_test: the number of axes for 3d vector is wrong"

    log, table, fig = V3.std_bound_test(limits = [0.2, 0.2, [-20, 10]], n_std = 3)
    assert len(log['data.x.x_1']['failed']) == 4, "std_bound_test: ['data.x.x_1']['failed'] for n_std = 3 is wrong"

    log, table, fig = V3.std_bound_test(limits = [0.2, 0.2, [-20, 10]], n_std = 1)
    assert len(log['data.x.x_1']['failed']) == 0, "std_bound_test: ['data.x.x_1']['failed'] for n_std = 1 is wrong"

    log, table, fig = V.std_bound_test(limits = 0.2, n_std = 3)
    assert len(log['data.x.x_1']['failed']) == 3, "std_bound_test: ['data.x.x_1']['failed'] for n_std = 3 is wrong"

    log, table, fig = V.std_bound_test(limits = [0.2], n_std = 1)
    assert len(log['data.x.x_1']['failed']) == 0, "std_bound_test: ['data.x.x_1']['failed'] for n_std = 1 is wrong"
    assert len(fig.axes) == 1, "std_bound_test: the number of axes for 1d vector is wrong"
    assert log['data.x.x_1']['passed'] == True, "std_bound_test: ['data.x.x_1']['passed'] is wrong"

    V3_bin = Validation(df, data_label = ['data.x.x_1', 'data.x.x_2','data.x.x_3'], param_label = 'data.time', 
                      method = 'bin', num = 50, units = 'm')
    log, table, fig = V3_bin.std_bound_test(limits = [0.25, 0.28, [-20, 10]], n_std = 1)
    assert log['data.x.x_1']['passed'] == True, "std_bound_test: nan_passed default True: log['data.x.x_1']['passed'] is wrong"
    assert list(log['data.x.x_1']['nan_std'].keys())[0] == 49, "std_bound_test: nan_passed default True: log['data.x.x_1']['nan_std']  is wrong"
    assert len(log['data.x.x_3']['failed']) == 49, "std_bound_test: nan_passed default True: log['data.x.x_3']['failed'] is wrong"

    log, table, fig = V3_bin.std_bound_test(limits = [0.25], n_std = 1, nan_passed = False)
    assert log['data.x.x_1']['passed'] == False, "std_bound_test: nan_passed False: log['data.x.x_1']['passed']  is wrong"
    assert list(log['data.x.x_1']['nan_std'].keys())[0] == 49, "std_bound_test: nan_passed = False: log['data.x.x_1']['nan_std'] is wrong"
    assert len(log['data.x.x_3']['failed']) == 50, "std_bound_test: nan_passed default True: log['data.x.x_3']['failed'] is wrong"

    plt.close('all')

    #nan+inf

    df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.y', 'data.y.y_1','data.y.y_2','data.y.y_3','data.time'])

    V = Validation(df, data_label = ['data.y.y_1'], param_label = 'data.y.y_3', 
                        method = 'bin', num = 50, units = 'm', omit_nan_rows = False)
    V3 = Validation(df, data_label = ['data.y.y_1', 'data.y.y_2'], param_label = 'data.y.y_3', 
                        method = 'scale', num = 50, units = 'm', omit_nan_rows = False)
    log3, table3, fig3 = V3.std_bound_test(limits = [30, [-25, 25]], n_std = 2)
    log, table, fig = V.std_bound_test(limits = 40, n_std = 3, connect_nan_std = True)
    passed = True
    for name in ['data.y.y_1', 'data.y.y_2']:
        passed = passed and log3[name]['passed']
    assert passed, 'std_bound_test, nan+inf 3d: omit_nan_rows False not passed'
    passed = True
    passed = passed and log['data.y.y_1']['passed']
    assert passed, 'std_bound_test, nan+inf 1d: omit_nan_rows False not passed'

    V = Validation(df, data_label = ['data.y.y_1'], param_label = 'data.y.y_3', 
                        method = 'scale', num = 50, units = 'm', omit_nan_rows = True)
    V3 = Validation(df, data_label = ['data.y.y_1', 'data.y.y_2'], param_label = 'data.y.y_3', 
                        method = 'bin', num = 50, units = 'm', omit_nan_rows = True)
    log3, table3, fig3 = V3.std_bound_test(limits = [30], n_std = 2, std_lines = False)
    log, table, fig = V.std_bound_test(limits = 40, n_std = 3, connect_nan_std = True, std_area = False)
    passed = True
    for name in ['data.y.y_1', 'data.y.y_2']:
        passed = passed and log3[name]['passed']
    assert passed, 'std_bound_test, nan+inf: omit_nan_rows True not passed'
    passed = True
    passed = passed and log['data.y.y_1']['passed']
    assert passed, 'std_bound_test, nan+inf: omit_nan_rows True not passed'

    plt.close('all')

def test_std_test():
    df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.x', 'data.x.x_1','data.x.x_2','data.x.x_3','data.time'])

    V = Validation(df, data_label = ['data.x.x_1'], param_label = 'data.time', 
                        method = 'bin', num = 50, units = 'm')
    V3 = Validation(df, data_label = ['data.x.x_1', 'data.x.x_2','data.x.x_3'], param_label = 'data.time', 
                        method = 'scale', num = 50, units = 'm')
    log, table, fig = V3.std_test(limits = [0.1, 0.2, 90], n_std = 1)
    
    assert (log['test_param']['limits'][0] - 0.1) < 0.001, "std_test: ['test_param']['limits'][0] is wrong"
    assert (log['test_param']['limits'][1] - 0.2) < 0.001, "std_test: ['test_param']['limits'][2][1] is wrong"
    assert log['data.x.x_1']['passed'] == True, "std_test: ['data.x.x_1']['passed'] is wrong"
    assert log['data.x.x_3']['passed'] == False, "std_test: ['data.x.x_3']['passed'] is wrong"
    assert abs(log['data.x.x_1']['pass_rate'] - 1.0) < 0.001, "std_test: ['data.x.x_1']['pass_rate'] is wrong"
    assert abs(log['data.x.x_3']['pass_rate'] - 0.96) < 0.001, "std_test: ['data.x.x_3']['pass_rate'] is wrong"

    assert all([a in log['data.x.x_3']['failed'].keys() for a in [13, 17]]) == True, "std_test: table['data.x.x_3'] is wrong"
    assert table['data.x.x_1'].all() == True, "std_test: table['data.x.x_1'] is wrong"
    assert len(fig.axes) == 3, "std_test: the number of axes for 3d vector is wrong"

    log, table, fig = V3.std_test(limits = 0.1, n_std = 3)
    assert len(log['data.x.x_3']['failed']) == 50, "std_test: ['data.x.x_3']['failed'] for n_std = 3 is wrong"

    log, table, fig = V.std_test(limits = 0.125, n_std = 2)
    assert len(log['data.x.x_1']['failed']) == 2, "std_test: ['data.x.x_1']['failed'] for n_std = 2 is wrong"

    log, table, fig = V.std_test(limits = [0.1], n_std = 1)
    assert len(log['data.x.x_1']['failed']) == 0, "std_test: ['data.x.x_1']['failed'] for n_std = 1 is wrong"
    assert len(fig.axes) == 1, "std_test: the number of axes for 1d vector is wrong"
    assert log['data.x.x_1']['passed'] == True, "std_test: ['data.x.x_1']['passed'] is wrong"

    V3_bin = Validation(df, data_label = ['data.x.x_1', 'data.x.x_2','data.x.x_3'], param_label = 'data.time', 
                      method = 'bin', num = 50, units = 'm')
    log, table, fig = V3_bin.std_test(limits = [0.1, 0.2, 90], n_std = 1)
    assert log['data.x.x_1']['passed'] == True, "std_test: nan_passed default True: log['data.x.x_1']['passed'] is wrong"
    assert list(log['data.x.x_1']['nan_std'].keys())[0] == 49, "std_test: nan_passed default True: log['data.x.x_1']['nan_std']  is wrong"
    assert len(log['data.x.x_3']['failed']) == 0, "std_test: nan_passed default True: log['data.x.x_3']['failed'] is wrong"

    log, table, fig = V3_bin.std_test(limits = [0.25], n_std = 1, nan_passed = False)
    assert log['data.x.x_3']['passed'] == False, "std_test: nan_passed False: log['data.x.x_3']['passed']  is wrong"
    assert list(log['data.x.x_1']['nan_std'].keys())[0] == 49, "std_test: nan_passed = False: log['data.x.x_1']['nan_std'] is wrong"
    assert len(log['data.x.x_3']['failed']) == 50, "std_test: nan_passed default True: log['data.x.x_3']['failed'] is wrong"

    plt.close('all')

    #nan+inf

    df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.y', 'data.y.y_1','data.y.y_2','data.y.y_3','data.time'])

    V = Validation(df, data_label = ['data.y.y_1'], param_label = 'data.y.y_3', 
                        method = 'bin', num = 50, units = 'm', omit_nan_rows = False)
    V3 = Validation(df, data_label = ['data.y.y_1', 'data.y.y_2'], param_label = 'data.y.y_3', 
                        method = 'scale', num = 50, units = 'm', omit_nan_rows = False)
    log3, table3, fig3 = V3.std_test(limits = [30, 25], n_std = 2)
    log, table, fig = V.std_test(limits = 40, n_std = 3, connect_nan_std = True)
    passed = True
    for name in ['data.y.y_1', 'data.y.y_2']:
        passed = passed and log3[name]['passed']
    assert passed, 'std_test, nan+inf 3d: omit_nan_rows False not passed'
    passed = True
    passed = passed and log['data.y.y_1']['passed']
    assert passed, 'std_test, nan+inf 1d: omit_nan_rows False not passed'

    V = Validation(df, data_label = ['data.y.y_1'], param_label = 'data.y.y_3', 
                        method = 'scale', num = 50, units = 'm', omit_nan_rows = True)
    V3 = Validation(df, data_label = ['data.y.y_1', 'data.y.y_2'], param_label = 'data.y.y_3', 
                        method = 'bin', num = 50, units = 'm', omit_nan_rows = True)
    log3, table3, fig3 = V3.std_test(limits = [30], n_std = 2, std_lines = False)
    log, table, fig = V.std_test(limits = 40, n_std = 3, connect_nan_std = True, std_area = False)
    passed = True
    for name in ['data.y.y_1', 'data.y.y_2']:
        passed = passed and log3[name]['passed']
    assert passed, 'std_bound_test, nan+inf: omit_nan_rows True not passed'
    passed = True
    passed = passed and log['data.y.y_1']['passed']
    assert passed, 'std_bound_test, nan+inf: omit_nan_rows True not passed'

    plt.close('all')

def test_mean_test():
    df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.x', 'data.x.x_1','data.x.x_2','data.x.x_3','data.time'])
    V = Validation(df, data_label = ['data.x.x_1'], param_label = 'data.time', 
                      method = 'bin', num = 50, units = 'm')
    V3 = Validation(df, data_label = ['data.x.x_1', 'data.x.x_2','data.x.x_3'], param_label = 'data.time', 
                        method = 'scale', num = 50, units = 'm')
    
    log, table, fig = V3.mean_test(limits = [[-0.1, 0.08], 0.1, 0.1])
    assert (log['test_param']['limits'][0][0] == -0.1) and (log['test_param']['limits'][0][1] == 0.08) \
        and (log['test_param']['limits'][1] == 0.1) and (log['test_param']['limits'][2] == 0.1), "mean_test: 'test_param' are wrong"
    
    assert log['data.x.x_1']['passed'] == True, "mean_test: ['data.x.x_1']['passed'] is wrong"
    assert abs(log['data.x.x_1']['pass_rate'] - 1.0) < 0.001, "mean_test: ['data.x.x_1']['pass_rate'] is wrong"
    assert log['data.x.x_3']['passed'] == False, "mean_test: ['data.x.x_3']['passed'] is wrong"
    assert abs(log['data.x.x_3']['pass_rate']) < 0.001, "mean_test: ['data.x.x_3']['pass_rate'] is wrong"
    assert len(log['data.x.x_3']['failed']) == 50, "mean_test: ['data.x.x_3']['failed'] is wrong"
    assert len(log['data.x.x_1']['failed']) == 0, "mean_test: ['data.x.x_1']['failed'] is wrong"
    assert len(fig.axes) == 3, "mean_test: the number of axes for 3d vector is wrong"

    log, table, fig = V3.mean_test(limits = [[-0.1, 0.05], 0.078])
    assert log is None, "mean_test: passing wrong limits does not lead to None output"

    log, table, fig = V3.mean_test(limits = [[-0.1, 0.08], 0.1, [0, 200]])
    assert all([x in ['data.time', 'data.x.x_1', 'data.x.x_2', 'data.x.x_3'] for x in table.columns]), "mean_test: columns are wrong"
    assert table[['data.x.x_1','data.x.x_2','data.x.x_3']].all().all() == True, "mean_test: table is wrong"

    log, table, fig = V3.mean_test(limits = [[-0.1, 0.08], 0.1, [0, 200]])
    assert all([x in ['data.time', 'data.x.x_1', 'data.x.x_2', 'data.x.x_3'] for x in table.columns]), "mean_test: columns are wrong"
    assert table[['data.x.x_1','data.x.x_2','data.x.x_3']].all().all() == True, "mean_test: table is wrong"

    log, table, fig = V.mean_test(limits = [-0.1, 0.08])
    # all pass the test
    assert len(fig.axes) == 1, "mean_test: the number of axes for 1d vector is wrong"
    assert log['data.x.x_1']['passed'] == True, "mean_test: 1d vector: ['data.x.x_1']['passed'] is wrong"
    assert abs(log['data.x.x_1']['pass_rate'] - 1) < 0.001, "mean_test: 1d vector: ['data.x.x_1']['pass_rate'] is wrong"
    assert len(log['data.x.x_1']['failed']) == 0, "mean_test: 1d vector: length of 'failed' is wrong"
    assert all([x in ['data.time', 'data.x.x_1'] for x in table.columns]) == True, "mean_test: 1d vector: table column names are wrong"

    log, table, fig = V.mean_test(limits = 0.001)
    # all fail test
    assert log['data.x.x_1']['passed'] == False, "mean_test: 1d vector: ['data.x.x_1']['passed'] is wrong"
    assert abs(log['data.x.x_1']['pass_rate']) < 0.001, "mean_test: 1d vector: ['data.x.x_1']['pass_rate'] is wrong"
    assert len(log['data.x.x_1']['failed']) == 50, "mean_test: 1d vector: length of 'failed' is wrong"

    log, table, fig = V.mean_test(limits = [0.001, 0.44, 0.22])
    assert log is None, "test_norm_test: passing wrong limits does not lead to None output"

    plt.close('all')

    #nan + inf
    df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.y', 'data.y.y_1','data.y.y_2','data.y.y_3','data.time'])
    V = Validation(df, data_label = ['data.y.y_1'], param_label = 'data.y.y_3', 
                        method = 'bin', num = 50, units = 'm', omit_nan_rows = False)
    V3 = Validation(df, data_label = ['data.y.y_1', 'data.y.y_2'], param_label = 'data.y.y_3', 
                        method = 'scale', num = 50, units = 'm', omit_nan_rows = False)
    log3, table3, fig3 = V3.mean_test(limits = [10])
    log, table, fig = V.mean_test(limits = [10])

    passed = True
    for name in ['data.y.y_1', 'data.y.y_2']:
        passed = passed and log3[name]['passed']
    assert passed, 'std_mean_test, nan+inf: omit_nan_rows False not passed'

    passed = log['data.y.y_1']['passed']
    assert passed, 'std_mean_test, nan+inf: omit_nan_rows False not passed'

    V = Validation(df, data_label = ['data.y.y_1'], param_label = 'data.y.y_3', 
                        method = 'scale', num = 50, units = 'm', omit_nan_rows = True)
    V3 = Validation(df, data_label = ['data.y.y_1', 'data.y.y_2'], param_label = 'data.y.y_3', 
                        method = 'bin', num = 50, units = 'm', omit_nan_rows = True)
    log3, table3, fig3 = V3.mean_test(limits = [10])
    log, table, fig = V.mean_test(limits = [10])

    passed = True
    for name in ['data.y.y_1', 'data.y.y_2']:
        passed = passed and log3[name]['passed']
    assert passed, 'std_mean_test, nan+inf: omit_nan_rows True not passed'

    passed = log['data.y.y_1']['passed']
    assert passed, 'std_mean_test, nan+inf: omit_nan_rows True not passed'

    plt.close('all')

def test_norm_test():
    df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.x', 'data.x.x_1','data.x.x_2','data.x.x_3','data.time'])
    V = Validation(df, data_label = ['data.x.x_1'], param_label = 'data.time', 
                      method = 'bin', num = 50, units = 'm')
    V3 = Validation(df, data_label = ['data.x.x_1', 'data.x.x_2','data.x.x_3'], param_label = 'data.time', 
                        method = 'scale', num = 50, units = 'm')
    
    log, table, fig = V3.norm_test(norm_type = 'Linf', limits = [[1.0, 0.1], 0.5])
    assert log is None, "test_norm_test: passing wrong limits does not lead to None output"

    log, table, fig = V3.norm_test(norm_type = 'Linf', limits = [1.0, 0.1, 0.5])
    assert log['data.x.x_1']['passed'] == True, "norm_test: ['data.x.x_1']['passed'] is wrong"
    assert log['data.x.x_3']['passed'] == False, "norm_test: ['data.x.x_1']['passed'] is wrong"
    assert abs(log['data.x.x_1']['pass_rate'] - 1.0) < 0.001, "norm_test: ['data.x.x_1']['pass_rate'] is wrong"
    assert abs(log['data.x.x_3']['pass_rate']) < 0.001, "norm_test: ['data.x.x_1']['pass_rate'] is wrong"
    assert len(log['data.x.x_3']['failed']) == 3, "number of failed sid is wrong"

    cols = ['data.x.x_1', 'data.x.x_2', 'data.x.x_3']
    for s in set(df['sid']):
        for col in cols:
            S = max(abs(V3.db.data.xs(s, level = 'sid')[col]))
            assert abs(S - log[col]['norm_value'][s]) < 0.0001, f"norm_test: sid = {s}, col = {col} is wrong"
    assert table['data.x.x_1'].all() and table['data.x.x_3'].any() == False, "norm_test Linf: table for is wrong"
    assert len(fig.axes) == 3, "norm_test: the number of axes for 3d vector is wrong"

    log, table, fig = V3.norm_test(norm_type = 'L2', limits = [1.0, 0.1, 0.5])
    cols = ['data.x.x_1', 'data.x.x_2', 'data.x.x_3']
    for s in set(df['sid']):
        for col in cols:
            S = 0
            for x in V3.db.data.xs(s, level = 'sid')[col]:
                S += x**2
            assert abs(np.sqrt(S) - log[col]['norm_value'][s]) < 0.0001, f"norm_test L2: sid = {s}, col = {col} is wrong"

    V = Validation(df, data_label = ['data.x.x_1'], param_label = 'data.time', 
                      method = 'bin', num = 50, units = 'm')
    log, table, fig = V.norm_test(norm_type = 'L2', limits = 0.4)
    assert log['data.x.x_1']['passed'] == True, "norm_test 1d L2: ['data.x.x_1']['passed'] is wrong"
    assert abs(log['data.x.x_1']['pass_rate'] - 1.0) < 0.001, "norm_test 1d L2: ['data.x.x_1']['pass_rate'] is wrong"
    assert len(log['data.x.x_1']['failed']) == 0, "norm tests 1d L2: number of failed sid is wrong"

    for s in set(df['sid']):
        S = 0
        for x in V.db.data.xs(s, level = 'sid')['data.x.x_1']:
            S += x**2
        assert abs(np.sqrt(S) - log['data.x.x_1']['norm_value'][s]) < 0.0001, f"norm_test 1d L2: sid = {s} is wrong"

    log, table, fig = V.norm_test(norm_type = 'Linf', limits = 0.1)
    assert log['data.x.x_1']['passed'] == True, "norm_test 1d Linf: ['data.x.x_1']['passed'] is wrong"
    assert abs(log['data.x.x_1']['pass_rate'] - 1.0) < 0.001, "norm_test 1d Linf: ['data.x.x_1']['pass_rate'] is wrong"
    assert len(log['data.x.x_1']['failed']) == 0, "norm_test 1d Linf: number of failed sid is wrong"

    for s in set(df['sid']):
        S = max(abs(V.db.data.xs(s, level = 'sid')['data.x.x_1']))
        assert abs(S - log['data.x.x_1']['norm_value'][s]) < 0.0001, f"norm_test 1d Linf: sid = {s} is wrong"

    plt.close('all')

    #nan+inf
    df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.y', 'data.y.y_1','data.y.y_2','data.y.y_3','data.time'])

    V = Validation(df, data_label = ['data.y.y_1'], param_label = 'data.y.y_3', 
                        method = 'scale', num = 50, units = 'm', omit_nan_rows = False)#, inf_vals = None)
    V3 = Validation(df, data_label = ['data.y.y_1', 'data.y.y_2'], param_label = 'data.y.y_3', 
                        method = 'bin', num = 50, units = 'm', omit_nan_rows = False)
    log3, table3, fig3 = V3.norm_test(norm_type = 'Linf', limits = 10)
    log, table, fig = V.norm_test(norm_type = 'Linf', limits = 1)

    passed = True
    for name in ['data.y.y_1', 'data.y.y_2']:
        passed = passed and log3[name]['passed']
    assert passed, 'norm_test 2d Linf, nan+inf: omit_nan_rows False not passed'
    passed = not log['data.y.y_1']['passed']
    assert passed, 'norm_test 1d Linf, nan+inf: omit_nan_rows False not passed'

    plt.close('all')

    V = Validation(df, data_label = ['data.y.y_1'], param_label = 'data.y.y_3', 
                        method = 'scale', num = 50, units = 'm', omit_nan_rows = True)#, inf_vals = None)
    V3 = Validation(df, data_label = ['data.y.y_1', 'data.y.y_2'], param_label = 'data.y.y_3', 
                        method = 'bin', num = 50, units = 'm', omit_nan_rows = True)
    log3, table3, fig3 = V3.norm_test(norm_type = 'Linf', limits = 10)
    log, table, fig = V.norm_test(norm_type = 'Linf', limits = 1)

    passed = True
    for name in ['data.y.y_1', 'data.y.y_2']:
        passed = passed and log3[name]['passed']
    assert passed, 'norm_test 2d Linf, nan+inf: omit_nan_rows False not passed'
    passed = not log['data.y.y_1']['passed']
    assert passed, 'norm_test 1d Linf, nan+inf: omit_nan_rows False not passed'

    plt.close('all')

    V = Validation(df, data_label = ['data.y.y_1'], param_label = 'data.y.y_3', 
                        method = 'scale', num = 50, units = 'm', omit_nan_rows = False)
    V3 = Validation(df, data_label = ['data.y.y_1', 'data.y.y_2'], param_label = 'data.y.y_3', 
                        method = 'bin', num = 50, units = 'm', omit_nan_rows =False)
    log3, table3, fig3 = V3.norm_test(norm_type = 'L2', limits = 35)
    log, table, fig = V.norm_test(norm_type = 'L2', limits = 25)

    passed = True
    for name in ['data.y.y_1', 'data.y.y_2']:
        passed = passed and log3[name]['passed']
    assert passed, 'norm_test 2d L2, nan+inf: omit_nan_rows False not passed'
    passed = not log['data.y.y_1']['passed']
    assert passed, 'norm_test 1d L2, nan+inf: omit_nan_rows False not passed'

    plt.close('all')

    V = Validation(df, data_label = ['data.y.y_1'], param_label = 'data.y.y_3', 
                            method = 'scale', num = 50, units = 'm', omit_nan_rows = True)
    V3 = Validation(df, data_label = ['data.y.y_1', 'data.y.y_2'], param_label = 'data.y.y_3', 
                        method = 'bin', num = 50, units = 'm', omit_nan_rows =True)
    log3, table3, fig3 = V3.norm_test(norm_type = 'L2', limits = 35)
    log, table, fig = V.norm_test(norm_type = 'L2', limits = 25)

    passed = True
    for name in ['data.y.y_1', 'data.y.y_2']:
        passed = passed and log3[name]['passed']
    assert passed, 'norm_test 2d L2, nan+inf: omit_nan_rows True not passed'
    passed = not log['data.y.y_1']['passed']
    assert passed, 'norm_test 1d L2, nan+inf: omit_nan_rows True not passed'

    plt.close('all')

def test_sid_test():
    df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.x', 'data.x.x_1','data.x.x_2','data.x.x_3','data.time'])
    V = Validation(df, data_label = ['data.x.x_1'], param_label = 'data.time', 
                      method = 'bin', num = 50, units = 'm')
    V3 = Validation(df, data_label = ['data.x.x_1', 'data.x.x_2','data.x.x_3'], param_label = 'data.time', 
                        method = 'scale', num = 50, units = 'm')
    log, table, fig = V3.sid_test(limits = [0.1, [-0.2, 0.02], 300])
    assert log['test_param']['limits'][0] == 0.1 and log['test_param']['limits'][1][0] == -0.2 \
        and log['test_param']['limits'][1][1] == 0.02 and log['test_param']['limits'][2] == 300, "sid test tests param are wrong"
    
    assert log['data.x.x_1']['passed'] == True, "sid test: ['data.x.x_1']['passed'] is wrong"
    assert log['data.x.x_2']['passed'] == False, "sid test: ['data.x.x_2']['passed'] is wrong"
    assert abs(log['data.x.x_1']['pass_rate']['sid_fraction'] - 1) < 0.001, "sid test: ['data.x.x_1']['passed'] is wrong"
    assert abs(log['data.x.x_2']['pass_rate']['sid_fraction']) < 0.001, "sid test: ['data.x.x_2']['passed'] is wrong"
    assert abs(log['data.x.x_1']['pass_rate'][3] - 1) < 0.001, "sid test: ['data.x.x_1']['passed'] is wrong"
    for s in set(table.index.get_level_values('sid')):
        assert abs(table['data.x.x_2'].xs(s, level = 'sid').value_counts()[True]/len(table['data.x.x_2'].xs(s, level = 'sid')) - \
                log['data.x.x_2']['pass_rate'][s]) < 0.001, f"sid test, sid = {s}: 'data.x.x_2' pass rate is wrong"
        assert abs(table['data.x.x_2'].xs(s, level = 'sid').value_counts()[False] - \
                len(log['data.x.x_2']['failed'][s])) < 0.001, f"sid test, sid = {s}: 'data.x.x_2' failed length is wrong"
    assert len(log['data.x.x_1']['failed']) == 0, "sid test, failed length is wrong"
    assert len(log['data.x.x_2']['failed']) == 3, "sid test, failed length is wrong"
    assert len(fig.axes) == 3, "sid test 3d vector, ax number is wrong"

    log, table, fig = V.sid_test(limits = [0.1, [-0.2, 0.02], 300])
    assert log is None, "passing wrong limits does not lead to None output"
    log, table, fig = V.sid_test(limits = [0.1])
    assert log['test_param']['limits'][0] == 0.1, "sid test 1d vector test_param is wrong"
    
    assert log['data.x.x_1']['passed'] == True, "sid test 1d: ['data.x.x_1']['passed'] is wrong"
    assert abs(log['data.x.x_1']['pass_rate']['sid_fraction'] - 1) < 0.001, "sid test: ['data.x.x_1']['pass_rate']['sid_fraction'] is wrong"
    assert abs(log['data.x.x_1']['pass_rate'][3] - 1) < 0.001, "sid test: ['data.x.x_1']['pass_rate'] sid 3 is wrong"
    assert len(log['data.x.x_1']['failed']) == 0, "sid test, failed length is wrong"

    log, table, fig = V.sid_test(limits = 0.01)
    assert abs(log['data.x.x_1']['pass_rate']['sid_fraction']) < 0.001, "sid test: ['data.x.x_1']['sid_fraction'] is wrong"
    for s in set(table.index.get_level_values('sid')):
        assert abs(table['data.x.x_1'].xs(s, level = 'sid').value_counts()[True]/len(table['data.x.x_1'].xs(s, level = 'sid')) - \
                log['data.x.x_1']['pass_rate'][s]) < 0.001, f"sid test, sid = {s}: 'data.x.x_1' pass rate is wrong"
        assert abs(table['data.x.x_1'].xs(s, level = 'sid').value_counts()[False] - \
                len(log['data.x.x_1']['failed'][s])) < 0.001, f"sid test, sid = {s}: 'data.x.x_1' failed length is wrong"
    assert len(log['data.x.x_1']['failed']) == 3, "sid test, failed length is wrong"

    plt.close('all')
    df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.y', 'data.y.y_1','data.y.y_2','data.y.y_3','data.time'])

    V = Validation(df, data_label = ['data.y.y_1'], param_label = 'data.y.y_3', 
                        method = 'scale', num = 50, units = 'm', omit_nan_rows = False)
    V3 = Validation(df, data_label = ['data.y.y_1', 'data.y.y_2'], param_label = 'data.y.y_3', 
                        method = 'bin', num = 50, units = 'm', omit_nan_rows =False)
    log3, table3, fig3 = V3.sid_test(limits = 10)
    log, table, fig = V.sid_test(limits = 10)

    passed = True
    for name in ['data.y.y_1', 'data.y.y_2']:
        passed = passed and log3[name]['passed']
    assert passed, 'norm_test 2d L2, nan+inf: omit_nan_rows False not passed'
    passed = log['data.y.y_1']['passed']
    assert passed, 'norm_test 1d L2, nan+inf: omit_nan_rows False not passed'

    V = Validation(df, data_label = ['data.y.y_1'], param_label = 'data.y.y_3', 
                            method = 'scale', num = 50, units = 'm', omit_nan_rows = True)
    V3 = Validation(df, data_label = ['data.y.y_1', 'data.y.y_2'], param_label = 'data.y.y_3', 
                        method = 'bin', num = 50, units = 'm', omit_nan_rows =True)
    log3, table3, fig3 = V3.sid_test(limits = 10)
    log, table, fig = V.sid_test(limits = 10)

    passed = True
    for name in ['data.y.y_1', 'data.y.y_2']:
        passed = passed and log3[name]['passed']
    assert passed, 'sid_test 2d, nan+inf: omit_nan_rows True not passed'
    passed = log['data.y.y_1']['passed']
    assert passed, 'sid_test 1d L2, nan+inf: omit_nan_rows True not passed'

    plt.close('all')

def test_set_tests():
    df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.x', 'data.x.x_1','data.x.x_2','data.x.x_3','data.time'])
    V = Validation(df, data_label = ['data.x.x_1'], param_label = 'data.time', 
                      method = 'bin', num = 50, units = 'm')
    V3 = Validation(df, data_label = ['data.x.x_1', 'data.x.x_2','data.x.x_3'], param_label = 'data.time', 
                        method = 'scale', num = 50, units = 'm')
    test_names = ['std_bound', 'mean', 'sid', 'norm_L2', 'norm_Linf']
    log, tables, figures = V3.set_tests(test_method = {'std_bound' : {'limits' : 1.0, 'n_std': 3, 'nan_passed': True},
                                       'mean' : {'limits' : 1.0},
                                       'sid' : {'limits' : 1.0},
                                       'norm_L2' : {'limits' : 1.0},
                                       'norm_Linf' : {'limits' : 1.0}})
    
    assert all([name in log.keys() for name in test_names]), "set_tests: test names are wrong"
    assert all([name in tables.keys() for name in test_names]), "set_tests: test names are wrong"
    assert all([name in figures.keys() for name in test_names]), "set_tests: test names are wrong"

    log, tables, figures = V.set_tests(test_method = {'std_bound' : {'limits' : 1.0, 'n_std': 3, 'nan_passed': True},
                                       'mean' : {'limits' : 1.0},
                                       'sid' : {'limits' : 1.0},
                                       'norm_L2' : {'limits' : 1.0},
                                       'norm_Linf' : {'limits' : 1.0}})
    
    assert all([name in log.keys() for name in test_names]), "set_tests 1d: test names are wrong"
    assert all([name in tables.keys() for name in test_names]), "set_tests 1d: test names are wrong"
    assert all([name in figures.keys() for name in test_names]), "set_tests 1d: test names are wrong"

    plt.close('all')

    #nan+inf
    df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.y', 'data.y.y_1','data.y.y_2','data.y.y_3','data.time'])
    V3 = Validation(df, data_label = ['data.y.y_1', 'data.y.y_2'], param_label = 'data.y.y_3', 
                        method = 'bin', num = 50, units = 'm', omit_nan_rows =False)
    log, tables, figures = V3.set_tests(test_method = {'std_bound' : {'limits' : 20.0, 'n_std': 1, 'nan_passed': True, 'std_area' : False, 'connect_nan_std':True},
                                        'mean' : {'limits' : 20.0},
                                        'sid' : {'limits' : 20.0},
                                        'norm_L2' : {'limits' : 35.0},
                                        'norm_Linf' : {'limits' : 10.0}})
    plt.close('all')
    passed = True
    for method in log.keys():
        for name in ['data.y.y_1', 'data.y.y_2']:
            passed = passed and log[method][name]['passed']
    assert passed, "set_tests, omit_nan_rows =False: not passed"

    V3 = Validation(df, data_label = ['data.y.y_1', 'data.y.y_2'], param_label = 'data.y.y_3', 
                        method = 'scale', num = 50, units = 'm', omit_nan_rows =True)
    log, tables, figures = V3.set_tests(test_method = {'std_bound' : {'limits' : 20.0, 'n_std': 1, 'nan_passed': True, 'std_area' : False, 'connect_nan_std':True},
                                        'mean' : {'limits' : 20.0},
                                        'sid' : {'limits' : 20.0},
                                        'norm_L2' : {'limits' : 40.0},
                                        'norm_Linf' : {'limits' : 10.0}})
    plt.close('all')
    passed = True
    for method in log.keys():
        for name in ['data.y.y_1', 'data.y.y_2']:
            passed = passed and log[method][name]['passed']
    assert passed, "set_tests, omit_nan_rows =True: not passed"


# ==============================================
#  ██████╗██╗████████╗██████╗  ██████╗ ███████╗
# ██╔════╝██║╚══██╔══╝██╔══██╗██╔═══██╗██╔════╝
# ██║     ██║   ██║   ██████╔╝██║   ██║███████╗
# ██║     ██║   ██║   ██╔══██╗██║   ██║╚════██║
# ╚██████╗██║   ██║   ██║  ██║╚██████╔╝███████║
#  ╚═════╝╚═╝   ╚═╝   ╚═╝  ╚═╝ ╚═════╝ ╚══════╝
# ==============================================


from .citros import Citros
from .citros_obj import (
    CitrosException,
    CitrosNotFoundException,
    FileNotFoundException,
    NoValidException,
)
from .utils import str_to_bool, suppress_ros_lan_traffic

from .batch import Batch
from .batch_uploader import NoConnectionToCITROSDBException

from .logger import get_logger, shutdown_log

from .service import data_access_service, NoDataFoundException

from .report import Report, NoNotebookFoundException

from .data import (
    CitrosDB,
    CitrosDict,
    CitrosData,
    CitrosDataArray,
    CitrosStat,
    Validation,
)

from .database import CitrosDB as CitrosDB_old

__all__ = [
    # citros
    "Citros",
    # citros_obj
    "CitrosException",
    "CitrosNotFoundException",
    "FileNotFoundException",
    "NoValidException",
    # utils
    "str_to_bool",
    "suppress_ros_lan_traffic",
    # batch
    "Batch",
    "NoConnectionToCITROSDBException",
    # logs
    "get_logger",
    "shutdown_log",
    # service
    "data_access_service",
    "NoDataFoundException",
    # reporting
    "Report",
    "NoNotebookFoundException",
    # data
    "CitrosDB",
    "CitrosDict",
    "CitrosData",
    "CitrosDataArray",
    "CitrosStat",
    "Validation",
    "CitrosDB_old",
]



import pandas as pd
from mpl_toolkits import mplot3d
from matplotlib.lines import Line2D
import matplotlib.pyplot as plt
import numpy as np

from citros.data.analysis import CitrosData

# from citros import CitrosData, CitrosDataArray, CitrosStat

from itertools import cycle

from ._utils import _get_logger


class _Plotter:
    def __init__(self, log=None):
        if log is None:            
            self.log = _get_logger(__name__)
        self.log = log

    def plot_graph(
        self,
        df,
        x_label,
        y_label,
        ax,
        legend,
        title,
        set_x_label,
        set_y_label,
        remove_nan,
        inf_vals,
        *args,
        **kwargs,
    ):
        """
        Plot graph '`y_label` vs. `x_label`' for each sid, where `x_label` and `y_label`
        are the labels of columns of the pandas.DataFrame `df`.

        Parameters
        ----------
        df : pandas.DataFrame
            Data table.
        x_label : str
            Label of the column to plot along x-axis.
        y_label : str
            Label of the column to plot along y-axis.
        *args : Any
            Additional arguments to style lines, set color, etc,
            see [matplotlib.axes.Axes.plot](https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.plot.html)
        ax : matplotlib.axes.Axes
            Figure axis to plot on. If not specified, the new pair of fig, ax will be created.
        legend : bool, default True
            If True, show the legend with sids.
        title : str
            Set title of the plot.
        set_x_label : str, default None
            Label to set to the x-axis. If None, label is set according to `x_label`.
        set_y_label : str, default None
            Label to set to the y-axis. If None, label is set according to `y_label`.
        remove_nan : bool, default True
            If True, NaN values will be removed before plotting.
        inf_vals : None or float, default 1e308
            If specified, all values that exceed the provided value in absolute terms will be removed before plotting.
            If this functionality is not required, set inf_vals = None.

        Returns
        -------
        fig : matplotlib.figure.Figure
            Created figure if `ax` is not passed.
        ax : matplotlib.axes.Axes
            Created axis if `ax` is not passed.

        Other Parameters
        ----------------
        **kwargs
            Other keyword arguments, see [matplotlib.axes.Axes.plot](https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.plot.html)
        """
        return_flag = False
        if ax is None:
            return_flag = True
            fig, ax = plt.subplots(figsize=(6, 6))

        if remove_nan:
            flag = df[x_label].notna() & df[y_label].notna()
        else:
            flag = pd.Series(data=[True] * len(df))
        if inf_vals is not None:
            flag = flag & (
                ((abs(df[x_label]) - inf_vals) < 0)
                & ((abs(df[y_label]) - inf_vals) < 0)
            )

        missing_col = False
        for col in ["sid", "rid"]:
            if col not in df.columns:
                missing_col = True
                self.log.error(f'column "{col}" must be in the DataFrame `df`')
        if missing_col:
            return None

        df_copy = df[flag].copy()

        try:
            df_copy.sort_values(by=["sid", "rid"], axis=0, inplace=True)
        except:
            pass
        df_copy.set_index("sid", inplace=True, drop=False)
        sid_list = list(set(df_copy.index))
        for s in sid_list:
            ax.plot(
                df_copy.loc[s][x_label],
                df_copy.loc[s][y_label],
                label=str(s),
                *args,
                **kwargs,
            )

        if set_x_label is None:
            ax.set_xlabel(x_label)
        else:
            ax.set_xlabel(set_x_label)
        if set_y_label is None:
            ax.set_ylabel(y_label)
        else:
            ax.set_ylabel(set_y_label)
        ax.grid()
        if title is not None:
            ax.set_title(title)
        if legend:
            ax.legend(title="sid", loc="best")

        if return_flag:
            return fig, ax
        else:
            return None

    def plot_3dgraph(
        self,
        df,
        x_label,
        y_label,
        z_label,
        ax,
        scale,
        legend,
        title,
        set_x_label,
        set_y_label,
        set_z_label,
        remove_nan,
        inf_vals,
        *args,
        **kwargs,
    ):
        """
        Plot 3D graph '`z_label` vs. `x_label` and `y_label`' for each sid, where `x_label`, `y_label` and `z_label`
        are the labels of columns of the pandas.DataFrame `df`.

        Parameters
        ----------
        df : pandas.DataFrame
            Data table.
        x_label : str
            Label of the column to plot along x-axis.
        y_label : str
            Label of the column to plot along y-axis.
        *args : Any
            Additional arguments to style lines, set color, etc,
            see [matplotlib.axes.Axes.plot](https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.plot.html)
        ax : matplotlib.axes.Axes
            Figure axis to plot on. If not specified, the new pair of fig, ax will be created.
        scale : bool, default True
            Specify whether the axis range should be the same for all axes.
        legend : bool, default True
            If True, show the legend with sids.
        title : str
            Set title of the plot.
        set_x_label : str, default None
            Label to set to the x-axis. If None, label is set according to `x_label`.
        set_y_label : str, default None
            Label to set to the y-axis. If None, label is set according to `y_label`.
        set_z_label : str, default None
            Label to set to the z-axis. If None, label is set according to `z_label`.
        remove_nan : bool, default True
            If True, NaN values will be removed before plotting.
        inf_vals : None or float, default 1e308
            If specified, all values that exceed the provided value in absolute terms will be removed before plotting.
            If this functionality is not required, set inf_vals = None.

        Returns
        -------
        fig : matplotlib.figure.Figure
            Created figure if `ax` is not passed.
        ax : matplotlib.axes.Axes
            Created axis if `ax` is not passed.

        Other Parameters
        ----------------
        **kwargs
            Other keyword arguments, see [matplotlib.axes.Axes.plot](https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.plot.html)
        """
        return_flag = False
        if ax is None:
            return_flag = True
            fig = plt.figure(figsize=(6, 6))
            ax = fig.add_subplot(111, projection="3d")

        if remove_nan:
            flag = df[x_label].notna() & df[y_label].notna() & df[z_label].notna()
        else:
            flag = pd.Series(data=[True] * len(df))
        if inf_vals is not None:
            flag = flag & (
                ((abs(df[x_label]) - inf_vals) < 0)
                & ((abs(df[y_label]) - inf_vals) < 0)
                & ((abs(df[z_label]) - inf_vals) < 0)
            )

        missing_col = False
        for col in ["sid", "rid"]:
            if col not in df.columns:
                missing_col = True
                self.log.error(f'column "{col}" must be in the DataFrame `df`')
        if missing_col:
            return None

        df_copy = df[flag].copy()

        try:
            df_copy.sort_values(by=["sid", "rid"], axis=0, inplace=True)
        except:
            pass
        df_copy.set_index("sid", inplace=True, drop=False)
        sid_list = list(set(df_copy.index))
        for s in sid_list:
            ax.plot(
                df_copy.loc[s][x_label],
                df_copy.loc[s][y_label],
                df_copy.loc[s][z_label],
                label=str(s),
                *args,
                **kwargs,
            )

        if set_x_label is None:
            ax.set_xlabel(x_label)
        else:
            ax.set_xlabel(set_x_label)
        if set_y_label is None:
            ax.set_ylabel(y_label)
        else:
            ax.set_ylabel(set_y_label)
        if set_z_label is None:
            ax.set_zlabel(z_label)
        else:
            ax.set_zlabel(set_z_label)
        if title is not None:
            ax.set_title(title)
        if legend:
            ax.legend(title="sid", loc="best")
        if scale:
            min_limit = min([df_copy[col].min() for col in [x_label, y_label, z_label]])
            max_limit = max([df_copy[col].max() for col in [x_label, y_label, z_label]])
            ax.axes.set_xlim3d(min_limit, max_limit)
            ax.axes.set_ylim3d(min_limit, max_limit)
            ax.axes.set_zlim3d(min_limit, max_limit)
        if return_flag:
            return fig, ax
        else:
            return None

    def multiple_y_plot(
        self,
        df,
        x_label,
        y_labels,
        fig,
        legend,
        title,
        set_x_label,
        set_y_label,
        remove_nan,
        inf_vals,
        *args,
        **kwargs,
    ):
        """
        Plot a series of vertically arranged graphs 'y vs. `x_label`', with the y-axis labels
        specified in the `y_labels` parameter.

        Different colors correspond to different sids.

        Parameters
        ----------
        df : pandas.DataFrame
            Data table.
        x_label : str
            Label of the column to plot along x-axis.
        y_labels : list of str
            Labels of the columns to plot along y-axis.
        *args : Any
            Additional arguments to style lines, set color, etc,
            see [matplotlib.axes.Axes.plot](https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.plot.html)
        fig : matplotlib.figure.Figure, optional
            If None, a new Figure will be created.
        legend : bool, default True
            If True, show the legend with sids.
        title : str
            Set title of the plot.
        set_x_label : str, default None
            Label to set to the x-axis. If None, label is set according to `x_label`.
        set_y_label : list of str, default None
            Labels to set to the y-axis. If None, label is set according to `y_labels`.
        remove_nan : bool, default True
            If True, NaN values will be removed before plotting.
        inf_vals : None or float, default 1e308
            If specified, all values that exceed the provided value in absolute terms will be removed before plotting.
            If this functionality is not required, set inf_vals = None.

        Returns
        -------
        fig : matplotlib.figure.Figure
            Created figure if `fig` is not passed.
        ax : numpy.ndarray of matplotlib.axes.Axes
            Created axis if `fig` is not passed.

        Other Parameters
        ----------------
        **kwargs
            Other keyword arguments, see [matplotlib.axes.Axes.plot](https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.plot.html)
        """
        if isinstance(y_labels, str):
            y_labels = [y_labels]
        if fig is None:
            return_flag = True
            fig, ax = plt.subplots(nrows=len(y_labels), ncols=1, figsize=(6, 6))
        else:
            return_flag = False
            ax = []
            for i in range(len(y_labels)):
                ax.append(fig.add_subplot(len(y_labels), 1, i + 1))
            if len(y_labels) == 1:
                ax = np.array([[ax]])
            else:
                ax = np.array(ax)

        missing_col = False
        for col in ["sid", "rid"]:
            if col not in df.columns:
                missing_col = True
                self.log.error(f'column "{col}" must be in the DataFrame `df`')
        if missing_col:
            return None

        df_copy = df.copy()
        try:
            df_copy.sort_values(by=["sid", "rid"], axis=0, inplace=True)
        except:
            pass

        sid_total_list = list(set(df_copy["sid"]))
        prop_cycle = plt.rcParams["axes.prop_cycle"]
        color_main_list = prop_cycle.by_key()["color"]
        if len(sid_total_list) <= len(color_main_list):
            color_dict = {
                sid_total_list[i]: color_main_list[i]
                for i in range(len(sid_total_list))
            }
        else:
            color_list = color_main_list + [
                tuple(np.random.uniform(0, 1, size=3))
                for _ in range(len(sid_total_list) - len(color_main_list))
            ]
            color_dict = {
                sid_total_list[i]: color_list[i] for i in range(len(color_list))
            }

        if set_y_label is None:
            set_y_label = y_labels
        else:
            if isinstance(set_y_label, str):
                set_y_label = [set_y_label]
            elif isinstance(set_y_label, list):
                if len(set_y_label) != len(y_labels):
                    self.log.warn(
                        f"the number of labels in 'set_y_label' to set to y-axes do not match the number of graphs.\nProvide {len(y_labels)} labels or set 'set_y_label' = None"
                    )
            else:
                set_y_label = y_labels

        for i, y_label in enumerate(y_labels):
            if remove_nan:
                flag = df_copy[y_label].notna() & df_copy[x_label].notna()
            else:
                flag = pd.Series(data=[True] * len(df_copy))
            if inf_vals is not None:
                flag = flag & (
                    ((abs(df_copy[x_label]) - inf_vals) < 0)
                    & ((abs(df_copy[y_label]) - inf_vals) < 0)
                )
            if x_label != y_label:
                F = df_copy[[x_label, y_label, "sid"]].loc[flag]
            else:
                F = df_copy[[x_label, "sid"]].loc[flag]
            F.set_index("sid", inplace=True, drop=False)
            sid_list = list(set(F.index))
            for s in sid_list:
                ax[i].plot(
                    F.loc[s][x_label],
                    F.loc[s][y_label],
                    label=str(s),
                    color=color_dict[s],
                    *args,
                    **kwargs,
                )
            ax[i].set_ylabel(set_y_label[i])
            ax[i].grid()
        if set_x_label is None:
            fig.supxlabel(x_label)
        else:
            fig.supxlabel(set_x_label)
        if title is not None:
            fig.suptitle(title)
        if legend:
            legend_dict = {}
            for ax_item in ax:
                handles, labels = ax_item.get_legend_handles_labels()
                legend_dict = {**legend_dict, **{l: h for l, h in zip(labels, handles)}}
            fig.legend(
                list(legend_dict.values()),
                list(legend_dict.keys()),
                bbox_to_anchor=(1.0, 0.94),
                loc="upper left",
                title="sid",
            )
        fig.tight_layout()
        if return_flag:
            return fig, ax

    def multiplot(
        self,
        df,
        labels,
        scale,
        fig,
        legend,
        title,
        set_x_label,
        set_y_label,
        remove_nan,
        inf_vals,
        label_all_xaxis,
        label_all_yaxis,
        num,
        *args,
        **kwargs,
    ):
        """
        Plot a matrix of N x N graphs, each displaying either the histogram with values distribution (for graphs on the diogonal) or
        the relationship between variables listed in `labels`, with N being the length of `labels` list.

        For non-diagonal graphs, colors are assigned to points according to sids.

        Parameters
        ----------
        df : pandas.DataFrame
            Data table.
        labels : list of str
            Labels of the columns to plot.
        *args : Any
            Additional arguments to style lines, set color, etc,
            see [matplotlib.axes.Axes.plot](https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.plot.html)
        scale : bool, default True
            Specify whether the axis range should be the same for x and y axes.
        fig : matplotlib.figure.Figure, optional
            If None, a new Figure will be created.
        legend : bool, default True
            If True, show the legend with sids.
        title : str
            Set title of the plot.
        set_x_label : list of str
            Labels to set to the x-axis. If None, label is set according to `labels`.
        set_y_label : list of str
            Labels to set to the y-axis. If None, label is set according to `labels`.
        remove_nan : bool, default True
            If True, NaN values will be removed before plotting.
        inf_vals : None or float, default 1e308
            If specified, all values that exceed the provided value in absolute terms will be removed before plotting.
            If this functionality is not required, set inf_vals = None.
        label_all_xaxis : bool, default False
            If True, x labels are set to the x-axes of the all graphs, otherwise only to the graphs in the bottom row.
        label_all_yaxis : bool, default False
            If True, y labels are set to the y-axes of the all graphs, otherwise only to the graphs in the first column.
        num : int, default 5
            Number of bins in the histogram on the diagonal.

        Returns
        -------
        fig : matplotlib.figure.Figure
            Created figure if `fig` is not passed.
        ax : numpy.ndarray of matplotlib.axes.Axes
            Created axis if `fig` is not passed.

        Other Parameters
        ----------------
        **kwargs
            Other keyword arguments, see [matplotlib.axes.Axes.plot](https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.plot.html)

        """
        if isinstance(labels, str):
            labels = [labels]
        N = len(labels)

        if fig is None:
            return_flag = True
            fig, axes = plt.subplots(nrows=N, ncols=N, figsize=(6, 6))
            if N == 1:
                axes = np.array([[axes]])
        else:
            return_flag = False
            axes = []
            for i in range(1, N * N + 1):
                axes.append(fig.add_subplot(N, N, i))
        axes = np.array(axes).reshape((N, N))

        missing_col = False
        for col in ["sid", "rid"]:
            if col not in df.columns:
                missing_col = True
                self.log.error(f'column "{col}" must be in the DataFrame `df`')
        if missing_col:
            return None

        df_copy = df.copy()
        try:
            df_copy.sort_values(by=["sid", "rid"], axis=0, inplace=True)
        except:
            pass

        sid_total_list = list(set(df_copy["sid"]))
        prop_cycle = plt.rcParams["axes.prop_cycle"]
        color_main_list = prop_cycle.by_key()["color"]
        if len(sid_total_list) <= len(color_main_list):
            color_dict = {
                sid_total_list[i]: color_main_list[i]
                for i in range(len(sid_total_list))
            }
        else:
            color_list = color_main_list + [
                tuple(np.random.uniform(0, 1, size=3))
                for _ in range(len(sid_total_list) - len(color_main_list))
            ]
            color_dict = {
                sid_total_list[i]: color_list[i] for i in range(len(color_list))
            }

        if set_x_label is None:
            set_x_label = labels
        else:
            if isinstance(set_x_label, str):
                set_x_label = [set_x_label]
            elif isinstance(set_x_label, list):
                if len(set_x_label) != N:
                    self.log.warn(
                        f"the number of labels in 'set_x_label' to set to x-axes do not match the number of graphs.\nProvide {N} labels or set 'set_x_label' = None"
                    )
            else:
                set_x_label = labels

        if set_y_label is None:
            set_y_label = labels
        else:
            if isinstance(set_y_label, str):
                set_y_label = [set_y_label]
            elif isinstance(set_y_label, list):
                if len(set_y_label) != N:
                    self.log.warn(
                        f"the number of labels in 'set_y_label' to set to y-axes do not match the number of graphs.\nProvide {N} labels or set 'set_y_label' = None"
                    )
            else:
                set_y_label = labels

        for i in range(N):
            y_label = labels[i]
            for j in range(N):
                x_label = labels[j]
                if i == j:
                    flag = df_copy[x_label].notna()
                    if inf_vals is not None:
                        flag = flag & ((abs(df_copy[x_label]) - inf_vals) < 0)
                    if x_label != "sid":
                        F = (
                            df_copy[[x_label, "sid"]]
                            .loc[flag]
                            .set_index("sid", drop=False)
                        )
                    else:
                        F = df_copy[["sid"]].loc[flag].set_index("sid", drop=False)
                    sid_list = list(set(F.index))
                    for s in sid_list:
                        axes[i][j].hist(
                            F[x_label].loc[s],
                            num,
                            histtype="step",
                            edgecolor=color_dict[s],
                        )
                else:
                    if remove_nan:
                        flag = df_copy[y_label].notna() & df_copy[x_label].notna()
                    else:
                        flag = pd.Series(data=[True] * len(df_copy))
                    if inf_vals is not None:
                        flag = flag & (
                            ((abs(df_copy[x_label]) - inf_vals) < 0)
                            & ((abs(df_copy[y_label]) - inf_vals) < 0)
                        )
                    if x_label != y_label:
                        F = df_copy[[x_label, y_label, "sid"]].loc[flag]
                    else:
                        F = df_copy[[x_label, "sid"]].loc[flag]
                    F = F.loc[:, ~F.columns.duplicated()]
                    F.set_index("sid", inplace=True, drop=False)
                    sid_list = list(set(F.index))
                    for s in sid_list:
                        axes[i][j].plot(
                            F.loc[s][x_label],
                            F.loc[s][y_label],
                            label=str(s),
                            color=color_dict[s],
                            *args,
                            **kwargs,
                        )
                    if scale:
                        min_limit = min([F[col].min() for col in [x_label, y_label]])
                        max_limit = max([F[col].max() for col in [x_label, y_label]])
                        axes[i][j].set_xlim(min_limit, max_limit)
                        axes[i][j].set_ylim(min_limit, max_limit)
                axes[i][j].grid()

                if label_all_xaxis:
                    axes[i][j].set_xlabel(set_x_label[j])
                else:
                    if i == (N - 1):
                        axes[i][j].set_xlabel(set_x_label[j])

                if label_all_yaxis:
                    axes[i][j].set_ylabel(set_y_label[i])
                else:
                    if j == 0:
                        axes[i][j].set_ylabel(set_y_label[i])

        if title is not None:
            fig.suptitle(title)
        if legend and N > 1:
            legend_dict = {}
            for i in range(0, N):
                for j in range(i + 1, N):
                    handles, labels = axes[i][j].get_legend_handles_labels()
                    legend_dict = {
                        **legend_dict,
                        **{l: h for l, h in zip(labels, handles)},
                    }
            fig.legend(
                list(legend_dict.values()),
                list(legend_dict.keys()),
                bbox_to_anchor=(1.0, 0.94),
                loc="upper left",
                title="sid",
            )
        fig.tight_layout()
        if return_flag:
            return fig, axes

    def plot_sigma_ellipse(
        self,
        df,
        x_label,
        y_label,
        ax,
        n_std,
        plot_origin,
        bounding_error,
        inf_vals,
        legend,
        title,
        set_x_label,
        set_y_label,
        scale,
        return_ellipse_param,
    ):
        """
        Plot sigma ellipses for the set of data.

        Parameters
        ----------
        df : pandas.DataFrame
            Data table.
        x_label : str
            Label of the column to plot along x-axis.
        y_labels : list of str
            Labels of the columns to plot along y-axis.
        ax : matplotlib.axes.Axes
            Figure axis to plot on. If not specified, the new pair of fig, ax will be created and returned.
        n_std : int or list of ints
            Radius of ellipses in sigmas.
        plot_origin: bool, default True
            If True, depicts origin (0, 0) with black cross.
        bounding_error : bool, default False
            If True, plots bounding error circle for each of the ellipses.
        inf_vals : None or float, default 1e308
            If specified, all values that exceed the provided value in absolute terms will be removed before plotting.
            If this functionality is not required, set inf_vals = None.
        legend : bool, default True
            If True, show the legend.
        title : str, optional
            Set title. If None, title is set as '`x_label` vs. `y_label`'.
        set_x_label : str, optional
            Set label of the x-axis. If None, label is set according to `x_label`.
        set_y_label : str, optional
            Set label of the y-axis. If None, label is set according to `y_label`.
        scale : bool, default False
            Specify whether the axis range should be the same for x and y axes.
        return_ellipse_param : bool, default False
            If True, returns ellipse parameters.

        Returns
        -------
        fig : matplotlib.figure.Figure
            Created figure if `ax` is not passed.
        ax : numpy.ndarray of matplotlib.axes.Axes
            Created axis if `ax` is not passed.
        ellipse_param : dict or list of dict
            Ellipse parameters if `return_ellipse_param` set True.<br />
            Parameters of the ellipse:
          - x : float
              x coordinate of the center.
          - y : float
              y coordinate of the center.
          - width : float
              Total ellipse width (diameter along the longer axis).
          - height : float
              Total ellipse height (diameter along the shorter axis).
          - alpha : float
              Angle of rotation, in degrees.<br />
            If bounding_error set True:
          - bounding_error : float
              Radius of the error circle.
        """
        if ax is None:
            return_flag = True
            fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(6, 6))

        flag = df[[x_label, y_label]].notna().all(axis=1)
        if inf_vals is not None:
            flag = (
                flag
                & ((df[x_label].abs() - inf_vals) < 0)
                & ((df[y_label].abs() - inf_vals) < 0)
            )
        x = df[flag][x_label].to_numpy(dtype=float)
        y = df[flag][y_label].to_numpy(dtype=float)
        if len(x) == 0:
            self.log.error("there is no data to plot")
            return
        if not isinstance(n_std, (list, np.ndarray)):
            n_std = [n_std]

        # plot points
        ax.plot(x, y, "k.")

        # plot ellipses
        color_list = cycle(
            ["tab:red", "tab:orange", "tab:olive", "tab:green", "tab:cyan", "tab:blue"]
        )
        ellipse_par_list = []
        if return_ellipse_param:
            ellipse_param = []
        if len(x) > 1:
            try:
                for n in n_std:
                    _, ellipse_par = CitrosData()._plot_ellipse(
                        x, y, n, ax, edgecolor=next(color_list)
                    )
                    ellipse_par_list.append(ellipse_par)
                    if return_ellipse_param:
                        ellipse_param.append(
                            {
                                "x": ellipse_par[3],
                                "y": ellipse_par[4],
                                "width": 2 * ellipse_par[0],
                                "height": 2 * ellipse_par[1],
                                "alpha": np.degrees(ellipse_par[2]),
                            }
                        )
                ax.plot(
                    ellipse_par[3], ellipse_par[4], "r+", mew=10, ms=2, label="mean"
                )

            except np.linalg.LinAlgError:
                self.log.error(
                    "can not calculate eigenvalues and eigenvectors of the covariance matrix to plot confidence ellipses"
                )
        else:
            bounding_error = False
            self.log.warn(
                "the number of points is not enough to plot confidence ellipses"
            )

        # plot center
        if plot_origin:
            ax.plot(0, 0, "k+", mew=10, ms=2, label="origin")

        # plot bounding error
        if bounding_error:
            R_bound_err = []
            y_text = 0
            for i, ellipse_par in enumerate(ellipse_par_list):
                R = CitrosData()._plot_bounding_error(*ellipse_par, ax)
                R_bound_err.append(R)
                ax.text(R, y_text, str(np.round(R, decimals=2)), fontsize=8)
                y_text = y_text - R / 20
                if return_ellipse_param:
                    ellipse_param[i]["bounding_error"] = R

        # set labels
        if set_x_label is None:
            ax.set_xlabel(x_label)
        else:
            ax.set_xlabel(set_x_label)
        if set_y_label is None:
            ax.set_ylabel(y_label)
        else:
            ax.set_ylabel(set_y_label)
        if title is None:
            ax.set_title(x_label + " vs. " + y_label)
        else:
            ax.set_title(title)
        if legend:
            fig.legend(bbox_to_anchor=(1.0, 0.96), loc="upper left")
        if scale:
            x_lim = ax.get_xlim()
            y_lim = ax.get_ylim()

            min_limit = min([x_lim[0], y_lim[0]])
            max_limit = max([x_lim[1], y_lim[1]])
            ax.set_xlim(min_limit, max_limit)
            ax.set_ylim(min_limit, max_limit)
        fig.tight_layout()
        ax.grid()
        if return_ellipse_param:
            if len(ellipse_param) == 1:
                ellipse_param = ellipse_param[0]
        if return_flag:
            if return_ellipse_param:
                return fig, ax, ellipse_param
            else:
                return fig, ax
        elif return_ellipse_param:
            return ellipse_param
        else:
            return None

    def time_plot(
        self,
        var_df,
        ax,
        var_name,
        sids,
        y_label,
        title_text,
        legend,
        *args,
        **kwargs,
    ):
        """
        Plot `var_name` vs. `Time` for each of the sids, where `Time` = `time_step` * rid.

        Parameters
        ----------
        var_df: pd.DataFrame
            pandas DataFrame
        ax : matplotlib.axes.Axes
            Figure axis to plot on.
        var_name : str
            Name of the variable to plot along y-axis.
        sids : list
            List of the sids. If specified, will override values that were set by `sid()` method.
            If not specified, data for all sids is used.
        y_label : str
            Label to set to y-axis. Default `var_name`.
        title_text : str
            Title of the figure. Default '`var_y_name` vs. Time'.
        legend : bool
            If True, show the legend with sids.
        *args : Any
            Additional arguments to style lines, set color, etc,
            see [matplotlib.axes.Axes.plot](https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.plot.html)

        Other Parameters
        ----------------
        **kwargs
            Other keyword arguments, see [matplotlib.axes.Axes.plot](https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.plot.html)
        """
        # plot the required simulations against time
        if sids is None or sids == []:
            sids = list(set(var_df.index))
        else:
            if isinstance(sids, int):
                sids = [sids]
            all_sids = list(set(var_df.index))
            bad_sids = []
            for s in sids:
                if s not in all_sids:
                    bad_sids.append(s)
            if len(bad_sids) != 0:
                self.log.warn("sids " + str(bad_sids) + " do not exist")
                sids = [s for s in sids if s not in bad_sids]
        for s in sids:
            ax.plot(
                var_df["Time"].loc[s], var_df[var_name].loc[s], label=s, *args, **kwargs
            )

        # add utilities to the plot
        ax.grid()
        if legend:
            ax.legend(title="sid", loc="best")
        if y_label is None:
            y_label = var_name
        ax.set_ylabel(y_label)
        ax.set_xlabel("Time [sec]")
        if title_text is None:
            title_text = var_name + " vs. Time"
        ax.set_title(title_text)

    def xy_plot(
        self,
        xy_df,
        ax,
        var_x_name,
        var_y_name,
        sids,
        x_label,
        y_label,
        title_text,
        legend,
        *args,
        **kwargs,
    ):
        """
        Plot `var_y_name` vs. `var_x_name` for each of the sids.

        Parameters
        ----------
        xy_df: pd.DataFrame
            pandas DataFrame
        ax : matplotlib.axes.Axes
            Figure axis to plot on.
        var_x_name : str
            Name of the variable to plot along x-axis.
        var_y_name : str
            Name of the variable to plot along y-axis.
        sids : int or list of int, optional
            List of the sids. If specified, will override values that were set by `sid()` method.
            If not specified, data for all sids is used.
        x_label : str, optional
            Label to set to x-axis. Default `var_x_name`.
        y_label : str, optional
            Label to set to y-axis. Default `var_y_name`.
        title_text : str, optional
            Title of the figure. Default '`var_y_name` vs. `var_x_name`'.
        legend : bool
            If True, show the legend with sids.
        *args : Any
            Additional arguments to style lines, set color, etc,
            see [matplotlib.axes.Axes.plot](https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.plot.html)

        Other Parameters
        ----------------
        **kwargs
            Other keyword arguments, see [matplotlib.axes.Axes.plot](https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.plot.html)
        """
        if sids is None or sids == []:
            sids = list(set(xy_df.index))
        else:
            if isinstance(sids, int):
                sids = [sids]
            all_sids = list(set(xy_df.index))
            bad_sids = []
            for s in sids:
                if s not in all_sids:
                    bad_sids.append(s)
            if len(bad_sids) != 0:
                self.log.warn("sids " + str(bad_sids) + " do not exist")
                sids = [s for s in sids if s not in bad_sids]

        for s in sids:
            ax.plot(
                xy_df[var_x_name].loc[s],
                xy_df[var_y_name].loc[s],
                label=s,
                *args,
                **kwargs,
            )
        if legend:
            ax.legend(title="sid", loc="best")
        if y_label is None:
            y_label = var_y_name
        ax.set_ylabel(y_label)
        if x_label is None:
            x_label = var_x_name
        ax.set_xlabel(x_label)
        if title_text is None:
            title_text = var_y_name + " vs. " + var_x_name
        ax.set_title(title_text)
        ax.grid()



import pytest
import random
import os
import tempfile
from faker import Faker
from citros import Citros

NUM_TEST_RUNS = 10

faker = Faker()


def generate_random_version():
    return f"{random.randint(1, 10)}.{random.randint(0, 9)}.{random.randint(0, 9)}"


def valid_setup_py_content():
    package_name = f"my_package_{faker.pystr(min_chars=5, max_chars=20)}"
    maintainer = faker.name()
    maintainer_email = faker.ascii_free_email()
    version = generate_random_version()
    description = faker.sentence()
    content = f"""
import os
from glob import glob
from setuptools import setup

package_name = '{package_name}'

setup(
    name=package_name,
    version='{str(version)}',
    packages=[package_name],
    install_requires=['setuptools'],
    zip_safe=True,
    author='ROS 2 Developer',
    author_email='foo@bar.com',
    maintainer='{maintainer}',
    maintainer_email='{maintainer_email}',
    keywords=['foo', 'bar'],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: TODO',
        'Programming Language :: Python',
        'Topic :: Software Development',
    ],
    description='{description}',
    license='TODO',
    entry_points={{
        'console_scripts': [
            'my_script = {package_name}.my_script:main'
        ],
    }},
)
"""
    return content, maintainer, maintainer_email, version, description


def generate_setup_py():
    for _ in range(NUM_TEST_RUNS):
        yield valid_setup_py_content()


@pytest.mark.parametrize(
    "valid_setup_py_content, maintainer, maintainer_email, version, description",
    generate_setup_py(),
)
def test_parse_setup_py(
    valid_setup_py_content, maintainer, maintainer_email, version, description
):
    with tempfile.TemporaryDirectory() as tmpdir:
        setup_path = os.path.join(tmpdir, "setup.py")
        with open(setup_path, "w") as f:
            f.write(valid_setup_py_content)
        with Citros() as citros:
            result = citros.parser_ros2.parse_setup_py(tmpdir)

        assert result["setup_py"] == setup_path
        assert result["version"] == str(version)
        assert result["maintainer"] == maintainer
        assert result["maintainer_email"] == maintainer_email
        assert result["description"] == description
